
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>

#include <netlink/netlink.h>
#include <netlink/socket.h>
#include <netlink/genl/ctrl.h>
#include <netlink/genl/genl.h>

#include <netlink/attr.h>

#include <json-c/json.h>

#include "config.h"

#define WPS_SUCCESS_TIME 5
#define NETLINK_FAMILY_NAME "easysoc"
#define NETLINK_GROUP_NAME  "notify"
#define SWITCH		"eth0"

static int switch_config_exists = 0;

/* attributes */
enum {
	NL_UNSPEC,
	NL_MSG,
	__NL_MAX,
};
#define NL_MAX (__NL_MAX - 1)
#define MAX_MSG 512

timer_t wps_timer = NULL;

static struct nla_policy nl_policy[__NL_MAX] = {
	[NL_MSG] = { .type = NLA_STRING },
};

static struct nlattr *attrs[__NL_MAX];

static void
remove_char(char *buf, char ch)
{
	char newbuf[strlen(buf)+1];
	int i = 0;
	int j = 0;

	while (buf[i]) {
		newbuf[j] = buf[i];
		if (buf[i] != ch)
			j++;
		i++;
	}
	newbuf[j] = '\0';
	strcpy(buf, newbuf);
}

static void
json_get_var(json_object *obj, const char *var, char *value)
{
	json_object_object_foreach(obj, key, val) {
		if(!strcmp(key, var)) {
			switch (json_object_get_type(val)) {
			case json_type_object:
				break;
			case json_type_array:
				break;
			case json_type_string:
				sprintf(value, "%s", json_object_get_string(val));
				break;
			case json_type_boolean:
				sprintf(value, "%d", json_object_get_boolean(val));
				break;
			case json_type_int:
				sprintf(value, "%d", json_object_get_int(val));
				break;
			default:
				break;
			}
		}
	}
}

static void
json_parse_and_get(char *str, const char *var, char *value)
{
	json_object *obj;

	obj = json_tokener_parse(str);
	if (is_error(obj) || json_object_get_type(obj) != json_type_object) {
		return;
	}
	json_get_var(obj, var, value);
}

static void signal_callback(int signal)
{
	system("ubus call led.wps set '{\"state\":\"off\"}'");
}

static void set_time(time_t time)
{
	struct itimerspec value;

	value.it_value.tv_sec = time;
	value.it_value.tv_nsec = 0;

	value.it_interval.tv_sec = 0;
	value.it_interval.tv_nsec = 0;

	timer_settime(wps_timer, 0, &value, NULL);
}

static void check_db_switch_config(void)
{
	FILE *hwsw;
	char line[128];

	if ((hwsw = popen("db -q get hw.switch", "r"))) {
		if(fgets(line, sizeof(line), hwsw) != NULL && strstr(line, "switch"))
			switch_config_exists = 1;
		pclose(hwsw);
	}
}

static int process_netlink_message(char *msg)
{
	char event[32];
	char data[128];
	char state[16];
	char port[16];
	char link[16];
	char mode[16];
	char status[32];
	char cmd[MAX_MSG];

	sscanf(msg, "%s %[^\n]s", event, data);

	if (strcmp(event, "wifi.wps") == 0) {
		if (wps_timer == NULL)
			timer_create(CLOCK_REALTIME, NULL, &wps_timer);

		set_time(0); // Reset the timer

		remove_char(data, '\'');
		json_parse_and_get(data, "state", state);
		if(strcmp(state, "off") == 0)
			system("ubus call led.wps set '{\"state\":\"off\"}'");
		else if (strcmp(state, "in process") == 0)
			system("ubus call led.wps set '{\"state\":\"notice\"}'");
		else if (strcmp(state, "overlap") == 0)
			system("ubus call led.wps set '{\"state\":\"alert\"}'");
		else if (strcmp(state, "error") == 0)
			system("ubus call led.wps set '{\"state\":\"alert\"}'");
		else if (strcmp(state, "success") == 0) {
			set_time(WPS_SUCCESS_TIME);
			system("ubus call router.wps stop");
			system("ubus call led.wps set '{\"state\":\"ok\"}'");
		}
	} else if (strcmp(event, "wifi.credentials") == 0) {
		snprintf(cmd, MAX_MSG, "wifi import %s", data);
		system(cmd);
	} else if (strcmp(event, "switch") == 0) {
		remove_char(data, '\'');
		json_parse_and_get(data, "port", port);
		json_parse_and_get(data, "link", link);

		if(strcmp(link, "up") == 0)
			strcpy(link, "add");
		else if (strcmp(link, "down") == 0)
			strcpy(link, "remove");

		if (switch_config_exists) {
			if(strcmp(port, "0") == 0)
				strcpy(port, "2");
			else if (strcmp(port, "1") == 0)
				strcpy(port, "1");

			snprintf(cmd, MAX_MSG, "ACTION=%s INTERFACE=%s.%s /sbin/hotplug-call net", link, SWITCH, port);
		} else {
			snprintf(cmd, MAX_MSG, "ACTION=%s INTERFACE=%s /sbin/hotplug-call switch", link, port);
		}
		system(cmd);
	} else if (strcmp(event, "dsl") == 0) {
		remove_char(data, '\'');
		json_parse_and_get(data, "mode", mode);
		json_parse_and_get(data, "status", status);

		snprintf(cmd, MAX_MSG, "STATUS=%s MODE=%s /sbin/hotplug-call dsl", status, mode);

		system(cmd);
	}

	return 0;
}

static int netlink_nl_parser(struct nl_msg *msg, void *arg)
{
	struct nlmsghdr *nlh = nlmsg_hdr(msg);
	char *data;
	int ret;
	char cmd[MAX_MSG];

	if (!genlmsg_valid_hdr(nlh, 0)){
		printf("Got invalid message\n");
		return 0;
	}

	ret = genlmsg_parse(nlh, 0, attrs, NL_MSG, nl_policy);

	if (!ret){

		if (attrs[NL_MSG] ) {
			data = nla_get_string(attrs[NL_MSG]);
			//printf("got message string (%s)\n", data);
			snprintf(cmd, MAX_MSG, "ubus send %s\n", data);
			system(cmd);
			process_netlink_message(data);
		}
	}
	return 0;
}

int main (int c, char* argv[])
{
	struct nl_sock *sock;
	int err;
	static int grp;

	sock = nl_socket_alloc();
	if(!sock){
		printf("Error: could not allocate socket\n");
		exit(1);
	}

	err = nl_socket_modify_cb(sock, NL_CB_VALID, NL_CB_CUSTOM, netlink_nl_parser, NULL);

	if ((err = genl_connect(sock)) < 0){
		printf("Error: %s\n", nl_geterror(err));
		exit(1);
	}

	if ((grp = genl_ctrl_resolve_grp(sock, NETLINK_FAMILY_NAME, NETLINK_GROUP_NAME)) < 0){
		printf("Error: %s (%s grp %s)\n", nl_geterror(grp), NETLINK_FAMILY_NAME, NETLINK_GROUP_NAME);
		exit(1);
	}

	nl_socket_add_membership(sock, grp);

	/* seq number dont really work for multicast so dissable check */
	nl_socket_disable_seq_check(sock);

		(void) signal(SIGALRM, signal_callback);


	check_db_switch_config();

	while (1) {
		err = nl_recvmsgs_default(sock);
		if (err < 0){
			printf("Error: %s (%s grp %s)\n", nl_geterror(err), NETLINK_FAMILY_NAME, NETLINK_GROUP_NAME);
		}
	}
}
